#!/usr/bin/perl -w


use strict;
use IO::File;
use POSIX;
use Fcntl ':flock';

my $gfile = '/etc/group.cued';
my $pfile = '/etc/passwd.cued';
my $sfile = '/etc/shadow.cued';
my $cfile = '/etc/crsmap';

my $sshkey = '/var/lib/passwd.cued/getpwfiles2_rsa';
my $lastupdatefile = '/var/lib/passwd.cued/lastupdate';
my $lockfile = '/var/lib/passwd.cued/getpwfiles.lockfile';

my $protodir = '/etc';
my $proto = 'passwd.proto';
my $gproto = 'group.proto';

my $debug = 0;
my $verbose = 0;
my $force = 0;

$SIG{'ALRM'} = \&timeout;

my $alarmdelay = 300;
alarm ($alarmdelay);

umask (022);

foreach my $arg (@ARGV) {
	$verbose = 1 if $arg eq '-v';
	$debug = 1 if $arg eq '-d';
	$force = 1 if $arg eq '-f';
}

print "Debug mode\n" if $debug;
print "Verbose mode\n" if $verbose;
print "Force update\n" if $force;

open (LF, ">$lockfile") or die "Unable to write $lockfile: $!\n";
flock (LF, (LOCK_EX | LOCK_NB)) or die "Unable to flock $lockfile: $!\n";
print "Locked\n" if $debug;

my $lastupdate = 0;
if ( -f $lastupdatefile) {
	open (LU, $lastupdatefile) 
		or die "Unable to read $lastupdatefile: $!\n";
	$lastupdate = <LU>;
	chomp $lastupdate;
	print "Lastupdate $lastupdate\n" if $debug;
	close (LU);
}

my $RAplines = parseproto ($proto);
foreach (@$RAplines) {
	print "$_\n" if $debug;
}
# only generate groups if there is a group.proto file
my $RAgplines;
if ( -f "$protodir/$gproto" ) {
	$RAgplines = parseproto ($gproto);
	foreach (@$RAgplines) {
		print "$_\n" if $debug;
	}
}

my (@pwdata, @gdata, @cdata);
my $mapfile;
my %mapdata;
my $mflag=0;
my $pflag=0;
my $gflag=0;
my $cflag=0;
my $lastchange=0;
my $eod=0;
open (SSH, "/usr/bin/ssh -n -T -oStrictHostKeyChecking=no -i $sshkey -q passwdd\@pw-serv|")
	or die "Cannot open ssh to pw-serv-vm: $!\n";
while (<SSH>) {
	chomp;
	if (/---lastchange (\d+)/) {
		$lastchange = $1;
		print "Lastchange = $lastchange\n" if $debug;
		next;
	}
	if (/---mapfile:(\S+)---/) {
		$mapfile = $1;
		$mflag=1;
		$pflag=0;
		$gflag=0;
		$cflag=0;
		next;	
	}
	if (/---passwd/) {
		$mflag=0;
		$pflag=1;
		$gflag=0;
		$cflag=0;
		next;
	}
	if (/---group/) {
		$mflag=0;
		$pflag=0;
		$gflag=1;
		$cflag=0;
		next;
	}
	if (/---crsmap/) {
		$mflag=0;
		$pflag=0;
		$gflag=0;
		$cflag=1;
		next;
	}
	if (/--eod/) {
		$eod=1;
		last;
	}
	push @{$mapdata{$mapfile}}, $_ if $mflag;
	push @pwdata, $_ if $pflag;
	push @gdata, $_ if $gflag;	
	push @cdata, $_ if $cflag;
	print "$_\n" unless ($mflag || $pflag || $gflag || $cflag);
	if (/^Failed to obtain lock/) {
		close (SSH);
		die "Server failed to get lock\n";
	}
}
close (SSH);
die "No End of Data marker read - aborting\n" unless $eod;

unless ($force || ($lastchange && $lastchange > $lastupdate)) {
	print "\nCurrent files are up to date\n" if $debug || $verbose;
	exit 0 unless $debug;
}

# Generate crsid map for our own purposes
my %crsmap;
foreach (@cdata) {
	my ($engid, $crsid) = split ' ';
	die "Bad crsmap line $_\n" unless $crsid;
	next if $engid eq $crsid;
	$crsmap{$engid}=$crsid;
}

foreach my $engid (sort keys %crsmap) {
	print "$engid -> $crsmap{$engid}\n" if $debug;
}

open (PWOUT, ">$pfile.new") or die "Unable to write $pfile.new: $!\n";
# Create new shadow file with correct permissions from the start
sysopen(SOUT, "$sfile.new", O_WRONLY|O_TRUNC|O_CREAT, 0000)
	or die "Unable to write $sfile.new: $!\n";
my $nplines = 0;
my %pwhash;
foreach my $pwline (@pwdata) {
	my ($name, $pw, $uid, $gid, $gcos, $hdir, $shell) =
		split (":", $pwline);
	print "Parsing $name from $pwline\n" if $debug;
	print "$name, $gcos\n" if $debug;
	$pwhash{$name} = { 
			pw => $pw,
			uid => $uid,
			gid => $gid,
			gcos => $gcos,
			hdir => $hdir,
			shell => $shell,
			};
}

# Scan for specific passwd.proto lines first, print specific users and mark.
# Look for first default line and record that too.
print "Outputting password lines\n" if $debug;
foreach my $pline (@$RAplines) {
	print $pline, "\n" if $debug;
	my ($name, $pw, $uid, $gid, $gcos, $hdir, $shell) =
		split (":", $pline);
	die "Bad pline $pline" unless defined $shell;
	if ($name =~ /^\+(.+)$/) {
		$name = $1;
		my $crsid = $crsmap{$name};
		print "Mapping to $crsid\n" if $crsid && $debug;
		$name = $crsid if $crsid;
		print "Fill in for $name\n" if $debug;
		unless (defined $pwhash{$name}) {
			warn "Unable to complete $name\n" 
				if ($debug || $verbose);
			next;
		}
		# de-dup in case both engid and crsid lines asked for
		# by proto file (which both map to crsid now)
		next if $pwhash{$name}{'output'};
		# Fill in fields which aren't already specified by proto
		$pw = $pwhash{$name}{'pw'} unless $pw;
		$uid = $pwhash{$name}{'uid'} unless $uid;
		$gid = $pwhash{$name}{'gid'} unless $gid;
		$gcos = $pwhash{$name}{'gcos'} unless $gcos;
		$hdir = $pwhash{$name}{'hdir'} unless $hdir;
		$shell = $pwhash{$name}{'shell'} unless $shell;
		$pwhash{$name}{'output'} = 1;
		print PWOUT "$name:x:$uid:$gid:$gcos:$hdir:$shell\n";
		print SOUT "$name:${pw}:::::::\n";
		$nplines++;
		next;
	}
	# Deal with default rules
	if ($name eq '+') {
		if (defined $pwhash{'default'}) {
			warn "Default pattern already defined\n";
			next;
		}
		$pwhash{'default'}{'seen'} = 1;	# make sure the hash is defined
		$pwhash{'default'}{'pw'} = $pw if $pw;
		$pwhash{'default'}{'uid'} = $uid if $uid;
		$pwhash{'default'}{'gid'} = $gid if $gid;
		$pwhash{'default'}{'gcos'} = $gcos if $gcos;
		$pwhash{'default'}{'hdir'} = $hdir if $hdir;
		$pwhash{'default'}{'shell'} = $shell if $shell;
		next;
	}
	# Non-interpolated proto lines:
	# Make sure fields are always filled.
	$pw = 'x' unless $pw;
	$uid = 0 unless $uid;
	$gid = 0 unless $gid;
	# Do homedir substitutions where appropriate
	if ($hdir =~ /\$/) {
		my $user = $name;	# user is a reasonable alias for name
		print "Interpolating $hdir\n" if $debug;
		$hdir = eval "\"$hdir\"" || $hdir; # revert if false
		print "result $hdir\n" if $debug;
		sleep 1 if $debug;
	}
	# Print result
	print PWOUT "$name:x:$uid:$gid:$gcos:$hdir:$shell\n";
	print SOUT "$name:${pw}:::::::\n";
	$pwhash{$name}{'output'} = 1;
	$nplines++;
}

# Now if a default is defined, print remaining users
if (defined $pwhash{'default'}) {
	print "Printing users matching default\n" if $debug;
#	foreach my $name (sort keys %pwhash) {
	# Rescan pwdata, so we preserve order of users from it
	foreach my $pwline (@pwdata) {
		my $name = (split (":", $pwline))[0];
		next if $pwhash{$name}{'output'};
		print "Default output $name\n" if $debug;
		my ($pw, $uid, $gid, $gcos, $hdir, $shell);
		$pw = $pwhash{'default'}{'pw'} || $pwhash{$name}{'pw'};
		$uid = $pwhash{'default'}{'uid'} || $pwhash{$name}{'uid'};
		$gid = $pwhash{'default'}{'gid'} || $pwhash{$name}{'gid'};
		$gcos = $pwhash{'default'}{'gcos'} || $pwhash{$name}{'gcos'};
		$hdir = $pwhash{'default'}{'hdir'} || $pwhash{$name}{'hdir'};
		$shell = $pwhash{'default'}{'shell'} || $pwhash{$name}{'shell'};
		# Make sure some fields are always filled.
		$pw = 'x' unless $pw;
		$uid = 0 unless $uid;
		$gid = 0 unless $gid;
		$gcos = '' unless $gcos;
		$hdir = '' unless $hdir;
		$shell = '' unless $shell;
		# Do homedir substitutions where appropriate
		if ($hdir =~ /\$/) {
			my $user = $name;# user is a reasonable alias for name
			print "Interpolating $hdir\n" if $debug;
			$hdir = eval "\"$hdir\"" || $hdir; # revert if false
			print "result $hdir\n" if $debug;
			sleep 1 if $debug;
		}
        	print PWOUT "$name:x:$uid:$gid:$gcos:$hdir:$shell\n";
		print SOUT "$name:${pw}:::::::\n";
		$nplines++;
	}
}
close (PWOUT) or die "unable to close $pfile.new: $!\n";
close (SOUT) or die "unable to close $sfile.new: $!\n";

my $plines = 0;
if (-f $pfile) {
	open (PW, $pfile) or die "Cannot read $pfile: $!\n";
	while (<PW>) {
		$plines++;
	}
	close (PW);
	if ($nplines + 100 < $plines) {
		warn "New password file more than 100 lines shorter\n";
		die "Exiting\n" unless $force;
	}
}
unless ($debug) {
	print "Installing $pfile\n" if $verbose;
	chmod (0644, "$pfile.new");
	rename ("$pfile.new", $pfile) 
		or die "Rename $pfile.new to $pfile failed: $!\n";
	print "Installing $sfile\n" if $verbose;
	chmod (0000, "$sfile.new");
	rename ("$sfile.new", $sfile) 
		or die "Rename $sfile.new to $sfile failed: $!\n";
}

open (GROUT, ">$gfile.new") or die "Unable to write $gfile.new: $!\n";
my %ghash;
foreach my $gline (@gdata) {
	my ($group, $star, $gid, $members) = split (":", $gline);
	$ghash{$group} = {
		gid => $gid,
		members => $members,
		};
}

my $nglines = 0;
# Same for group.proto file as for passwd.proto file
foreach my $gpline (@$RAgplines) {
	print $gpline, "\n" if $debug;
	my ($group, $star, $gid, $members) = split (":", $gpline);
	if ($group =~ /^\+(.+)$/) {
		$group = $1;
		unless (defined $ghash{$group}) {
			warn "Unable to complete $group\n"
				if ($debug || $verbose);
			next;
		}
		$gid = $ghash{$group}{'gid'} unless $gid;
		$members = $ghash{$group}{'members'} unless $members;
		print GROUT "$group:*:$gid:$members\n";
		$ghash{$group}{'output'} = 1;
		$nglines++;
		next;
	}
	if ($group eq '+') {
		print "Processing default line $gpline\n" if $debug;
		if (defined $ghash{'default'}) {
			warn "Default group pattern already defined\n";
			next;
		}
		$ghash{'default'}{'seen'} = 1;	# make sure the hash is defined
		$ghash{'default'}{'gid'} = $gid if $gid;
		$ghash{'default'}{'members'} = $members if $members;
		next;
	}
	# Non-interpolated proto lines:
	# Make sure some fields always filled
	$gid = 0 unless $gid;
	$members = '' unless $members;
	print GROUT "$group:*:$gid:$members\n";
	$nglines++;
}

if (defined $ghash{'default'}) {
	print "Printing groups matching default\n" if $debug;
	my ($gid, $members);
	foreach my $group (sort keys %ghash) {
		next if $group eq 'default';
                next if $ghash{$group}{'output'};
		$gid = $ghash{'default'}{'gid'} || $ghash{$group}{'gid'}
				|| 0;
		$members = $ghash{'default'}{'members'} 
				|| $ghash{$group}{'members'} || '';
		if ($members eq '') {
			print GROUT "$group:*:$gid:$members\n";
			$nglines++;
			next;
		} 
		my $count = 0;
		my $gline = "$group:*:$gid:";
		foreach my $member (split ",", $members) {
			$gline .= "$member,";
			$count++;
			if ($count > 9) {
				chop $gline;
				print GROUT "$gline\n";
				$nglines++;
				$gline = "$group:*:$gid:";
				$count = 0;
			}
		}
		chop $gline;
		print GROUT "$gline\n";
		$nglines++;
	}
}
close (GROUT) or die "Unable to close $gfile.new: $!\n";

my $glines = 0;
if ( -f $gfile ) {
	open (GRP, $gfile) or die "unable to read $gfile: $!\n";
	while (<GRP>) {
		$glines++;
	}
	close (GRP);
	if ($nglines + 100 < $glines) {
		warn "New group file more than 100 lines shorter\n";
		die "Exiting\n" unless $force;
	}
}
unless ($debug) {
	print "Installing $gfile\n" if $verbose;
	chmod (0644, "$gfile.new");
	rename ("$gfile.new", $gfile)
		or die "Rename $gfile.new to $gfile failed: $!\n";
}

foreach my $map (keys %mapdata) {
	my $mapfile = "/etc/$map";
	my $nmapfile = "$mapfile.new";
	print "Creating $mapfile.new\n" if $debug;
	open (NMAP, ">$nmapfile") or die "Unable to write $nmapfile: $!\n";
	print NMAP join "\n", @{$mapdata{$map}};
	print NMAP "\n";
	close (NMAP) or die "unable to close $mapfile.new: $!\n";
	unless ($debug) {
		print "Installing $mapfile\n" if $verbose;
		chmod (0644, $nmapfile);
		rename ($nmapfile, $mapfile) 
			or die "Rename $nmapfile to $mapfile failed: $!\n";
	}
}


# crsmap only generated if the file already exists
if (-f $cfile) {
	open (NCRS, ">$cfile.new") or die "Unable to write $cfile.new: $!\n";
	print NCRS join "\n", @cdata;
	print NCRS "\n";
	close (NCRS) or die "unable to close $cfile.new: $!\n";
	chmod (0644, "$cfile.new");

	unless ($debug) {
		print "Installing $cfile\n" if $verbose;
		rename ("$cfile.new", $cfile)
			or die "Rename $cfile.new to $cfile failed: $!\n";
	}
}

open (LU, ">$lastupdatefile.new")
	or die "Unable to write $lastupdatefile.new: $!\n";
print LU $lastchange;
close (LU);
chmod (0644, "$lastupdatefile.new");
unless ($debug) {
	print "Updating $lastupdatefile\n" if $verbose;
	rename ("$lastupdatefile.new", $lastupdatefile);
}
exit 0;

#
# Subroutines
#

sub parseproto {
	my ($protofile) = @_;
	my @protolines;
	open (my $pfh, "$protodir/$protofile") or die "Unable to read $protodir/$protofile: $!\n";
	print "Reading $protofile\n" if $debug;
	while (<$pfh>) {
		chomp;
		next if /^#/;
		if (/^%include (\S+)/) {
			my $incfile = $1;
			print "Including $protodir/$incfile\n" if $debug;
			my $RAinclines = parseproto ($incfile);
			@protolines = (@protolines, @$RAinclines);
			next;
		}
		push (@protolines, $_);
	}
	return \@protolines;
}

sub timeout {
	die "Alarm went off after $alarmdelay seconds\n";
}
